/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 *
 */

package ru.ulstu.method;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import ru.ulstu.TimeSeriesUtils;
import ru.ulstu.datamodel.Model;
import ru.ulstu.datamodel.exception.ForecastValidateException;
import ru.ulstu.datamodel.exception.ModelingException;
import ru.ulstu.datamodel.exception.TimeSeriesValidateException;
import ru.ulstu.datamodel.ts.TimeSeries;
import ru.ulstu.datamodel.ts.TimeSeriesValue;

import java.time.temporal.ChronoUnit;
import java.util.List;

/**
 * Наиболее общая логика моделировании и прогнозирования временных рядов
 */
public abstract class Method {
    @JsonIgnore
    public abstract List<MethodParameter> getAvailableParameters();

    /**
     * Возвращает модельное представление валидного временного ряда: для тех же точек времени что и в параметре timeSeries
     * строится модель. Количество точек может быть изменено: сокращено при сжатии ряда, увеличено при интерполяции.
     *
     * @return модель временного ряда
     */
    protected abstract Model getModelOfValidTimeSeries(TimeSeries timeSeries, List<MethodParamValue> parameters);

    /**
     * Возвращает модельное представление временного ряда: для тех же точек времени что и в параметре timeSeries
     * строится модель. Количество точек может быть изменено: сокращено при сжатии ряда, увеличено при интерполяции.
     * Метод является шаблонным, выполняет операции валидации исходного ряда и потом его моделирование
     * <p>
     * return модельное представление временного ряда
     *
     * @throws ModelingException генерируется, если есть проблемы моделирования при задании параметров
     */
    public Model getModel(TimeSeries timeSeries, List<MethodParamValue> parameters) throws ModelingException {
        validateTimeSeries(timeSeries);
        validateAdditionalParams(timeSeries, parameters);
        return getModelOfValidTimeSeries(timeSeries, parameters);
    }

    /**
     * Выполняет построение прогноза временного ряда. Даты спрогнозированных точек будут сгенерированы по модельным точкам.
     *
     * @param model    модель временного ряда, включающая все нужные компоненты
     * @param forecast заготовка временного ряда для прогноза с датами и нужным количеством точек для прогнозирования
     * @return прогноз временного ряда
     */
    protected abstract TimeSeries getForecastWithValidParams(Model model, TimeSeries forecast) throws ModelingException;

    public boolean canMakeForecast(TimeSeries timeSeries, List<MethodParamValue> parameters, int countPoints) {
        try {
            validateTimeSeries(timeSeries);
            validateAdditionalParams(timeSeries, parameters);
            validateForecastParams(countPoints);
        } catch (ModelingException ex) {
            return false;
        }
        return true;
    }

    public boolean canMakeModel(TimeSeries timeSeries, List<MethodParamValue> parameters) {
        try {
            validateTimeSeries(timeSeries);
            validateAdditionalParams(timeSeries, parameters);
        } catch (ModelingException ex) {
            return false;
        }
        return true;
    }

    /**
     * Выполняет построение модели и прогноза временного ряда. Даты спрогнозированных точек будут сгенерированы
     * по модельным точкам.
     *
     * @param countPoints количество точек для прогнозирования
     * @return прогноз временного ряда
     */
    public TimeSeries getForecast(TimeSeries timeSeries,
                                  List<MethodParamValue> parameters,
                                  int countPoints) throws ModelingException {
        validateForecastParams(countPoints);
        Model model = getModel(timeSeries, parameters);
        TimeSeries forecast = generateEmptyForecastPoints(model.getTimeSeriesModel(), countPoints);
        forecast.getFirstValue().setValue(model.getTimeSeriesModel().getLastValue().getValue());
        return getForecastWithValidParams(model, forecast);
    }

    protected TimeSeries generateEmptyForecastPoints(TimeSeries model, int countPointForecast) {
        long diffMilliseconds = TimeSeriesUtils.getTimeDifferenceInMilliseconds(model);
        TimeSeries forecast = new TimeSeries("Forecast of " + model.getName());
        forecast.addValue(new TimeSeriesValue(model.getLastValue().getDate()));
        for (int i = 1; i < countPointForecast + 1; i++) {
            forecast.addValue(new TimeSeriesValue(forecast.getValues().get(i - 1).getDate().plus(diffMilliseconds, ChronoUnit.MILLIS)));
        }
        return forecast;
    }

    private void validateForecastParams(int countPoints) throws ForecastValidateException {
        if (countPoints < 1) {
            throw new ForecastValidateException("Количество прогнозных точек должно быть больше 0");
        }
    }

    protected void validateTimeSeries(TimeSeries timeSeries) throws ModelingException {
        if (timeSeries == null || timeSeries.isEmpty()) {
            throw new TimeSeriesValidateException("Временной ряд должен быть не пустым");
        }
        if (timeSeries.getLength() < 2) {
            throw new TimeSeriesValidateException("Временной ряд должен содержать хотя бы 2 точки");
        }
        if (timeSeries.getValues().stream().anyMatch(val -> val == null || val.getValue() == null)) {
            throw new TimeSeriesValidateException("Временной ряд содержит пустые значения");
        }
        if (timeSeries.getValues().stream().anyMatch(val -> val.getDate() == null)) {
            throw new TimeSeriesValidateException("Временной ряд должен иметь отметки времени");
        }
    }

    protected void validateAdditionalParams(TimeSeries timeSeries, List<MethodParamValue> parameters) throws ModelingException {

    }

    @Override
    @JsonProperty("name")
    public String toString() {
        return getClass().getSimpleName();
    }
}
