/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 *
 */

package ru.ulstu.service;

import org.springframework.stereotype.Service;
import ru.ulstu.datamodel.Model;
import ru.ulstu.datamodel.ModelingResult;
import ru.ulstu.datamodel.exception.ModelingException;
import ru.ulstu.datamodel.ts.TimeSeries;
import ru.ulstu.datamodel.ts.TimeSeriesValue;
import ru.ulstu.method.Method;
import ru.ulstu.method.MethodParamValue;
import ru.ulstu.method.MethodParameter;
import ru.ulstu.score.ScoreMethod;
import ru.ulstu.score.Smape;

import java.lang.reflect.InvocationTargetException;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

@Service
public class MethodParamBruteForce {
    private final int DEFAULT_THREAD_COUNT = 50;
    private final List<Method> methods;
    private final ScoreMethod scoreMethod = new Smape();
    private final ExecutorService executors = Executors.newFixedThreadPool(DEFAULT_THREAD_COUNT);

    public MethodParamBruteForce(List<Method> methods) {
        this.methods = methods;
    }

    public ModelingResult getForecast(TimeSeries timeSeries, int countPointsForecast) throws ExecutionException, InterruptedException, NoSuchMethodException, InvocationTargetException, InstantiationException, IllegalAccessException, ModelingException {
        List<Future<ModelingResult>> results = new ArrayList<>();
        List<ModelingResult> results2 = new CopyOnWriteArrayList<>();
        final int countPoints = (countPointsForecast > timeSeries.getLength()) ? timeSeries.getLength() / 3 : countPointsForecast;
        TimeSeries reducedTimeSeries = new TimeSeries(timeSeries.getValues().stream().limit(timeSeries.getLength() - countPoints).collect(Collectors.toList()),
                "test part of " + timeSeries.getName());

        Map<LocalDateTime, Double> tsValues = timeSeries.getValues().stream()
                .collect(Collectors.toMap(TimeSeriesValue::getDate, TimeSeriesValue::getValue));

        for (Method method : methods) {
            List<List<MethodParamValue>> availableParametersValues = getAvailableParametersValues(method.getAvailableParameters());
            for (List<MethodParamValue> parametersValues : availableParametersValues) {
                Method methodInstance = method.getClass().getDeclaredConstructor().newInstance();
                if (methodInstance.canMakeForecast(reducedTimeSeries, parametersValues, countPoints)) {
                    results.add(executors.submit(() -> {
                        TimeSeries forecast = methodInstance.getForecast(reducedTimeSeries, parametersValues, countPoints);
                        return new ModelingResult(forecast,
                                parametersValues,
                                scoreMethod.getScore(tsValues, forecast),
                                methodInstance);
                    }));
                }
            }
        }
        for (Future<ModelingResult> futureModelingResult : results) {
            results2.add(futureModelingResult.get());
        }
        ModelingResult bestResult = results2.stream()
                .min(Comparator.comparing(modelingResult -> modelingResult.getScore().getDoubleValue()))
                .orElseThrow(() -> new ModelingException("Лучший метод не найден"));

        TimeSeries forecast = bestResult.getTimeSeriesMethod().getForecast(timeSeries,
                bestResult.getParamValues(),
                countPoints);
        forecast.getValue(0).setValue(timeSeries.getNumericValue(timeSeries.getLength() - 1));

        return new ModelingResult(forecast,
                bestResult.getParamValues(),
                bestResult.getScore(),
                bestResult.getTimeSeriesMethod());
    }

    public TimeSeries getForecastWithOptimalLength(TimeSeries timeSeries) {
        throw new RuntimeException("Not implemented");
    }

    public ModelingResult getSmoothedTimeSeries(TimeSeries timeSeries) throws ExecutionException, InterruptedException, NoSuchMethodException, InvocationTargetException, InstantiationException, IllegalAccessException {
        List<Future<ModelingResult>> results = new ArrayList<>();
        List<ModelingResult> results2 = new CopyOnWriteArrayList<>();

        Map<LocalDateTime, Double> tsValues = timeSeries.getValues().stream()
                .collect(Collectors.toMap(TimeSeriesValue::getDate, TimeSeriesValue::getValue));

        for (Method method : methods) {
            List<List<MethodParamValue>> availableParametersValues = getAvailableParametersValues(method.getAvailableParameters());
            for (List<MethodParamValue> parametersValues : availableParametersValues) {
                Method methodInstance = method.getClass().getDeclaredConstructor().newInstance();
                if (methodInstance.canMakeModel(timeSeries, parametersValues)) {
                    results.add(executors.submit(() -> {
                        Model model = methodInstance.getModel(timeSeries, parametersValues);
                        return new ModelingResult(model.getTimeSeriesModel(),
                                parametersValues,
                                scoreMethod.getScore(tsValues, model.getTimeSeriesModel()),
                                methodInstance);
                    }));
                }
            }
        }
        for (Future<ModelingResult> futureModelingResult : results) {
            results2.add(futureModelingResult.get());
        }
        return results2.stream()
                .min(Comparator.comparing(modelingResult -> modelingResult.getScore().getDoubleValue()))
                .orElse(null);
    }

    private List<List<MethodParamValue>> getAvailableParametersValues(List<MethodParameter> availableParameters) {
        List<List<MethodParamValue>> result = new ArrayList<>();
        Map<MethodParameter, Integer> parameterOffset = new TreeMap<>();
        Map<MethodParameter, List<Number>> parameterValues = new TreeMap<>();
        for (MethodParameter methodParameter : availableParameters) {
            parameterOffset.put(methodParameter, 0);
            parameterValues.put(methodParameter, methodParameter.getAvailableValues());
        }
        while (!isAllValuesUsed(parameterOffset, parameterValues)) {
            List<MethodParamValue> resultRow = new ArrayList<>();
            for (MethodParameter methodParameter : parameterOffset.keySet()) {
                resultRow.add(new MethodParamValue(methodParameter,
                        parameterValues.get(methodParameter).get(parameterOffset.get(methodParameter))));
            }
            incrementOffset(parameterOffset, parameterValues);
            result.add(resultRow);
        }
        return result;
    }

    private void incrementOffset(Map<MethodParameter, Integer> parameterOffset,
                                 Map<MethodParameter, List<Number>> parameterValues) {
        List<MethodParameter> parameters = new ArrayList<>(parameterOffset.keySet());
        int i = 0;
        while (i < parameters.size() && !isAllValuesUsed(parameterOffset, parameterValues)) {
            if (parameterOffset.get(parameters.get(i)) == parameterValues.get(parameters.get(i)).size() - 1) {
                parameterOffset.put(parameters.get(i), 0);
                i++;
                continue;
            }
            if (parameterOffset.get(parameters.get(i)) < parameterValues.get(parameters.get(i)).size()) {
                parameterOffset.put(parameters.get(i), parameterOffset.get(parameters.get(i)) + 1);
                return;
            }
            i++;
        }
    }

    private boolean isAllValuesUsed(Map<MethodParameter, Integer> parameterOffset,
                                    Map<MethodParameter, List<Number>> parameterValues) {
        for (MethodParameter methodParameter : parameterOffset.keySet()) {
            if (parameterOffset.get(methodParameter) != parameterValues.get(methodParameter).size() - 1) {
                return false;
            }
        }
        return true;
    }
}
